BeginPackage["HEPMath`Tensors`"]

(* declarations *)

IndexTypeQ::usage = "IndexTypeQ[i] returns True if i has been declared as an
index type with DeclareIndexType"

IndexDimension::usage = "IndexDimension[t] returns the dimension of index type
t, as declared in DeclareIndexType."

DeclareIndexType::usage = "DeclareIndexType[t, dim] declares type t as a new
index type with index dimension dim. DeclareIndexType[t, dim, id] additionally
declares id as the identity matrix in the space of indices of type t."

DeclareIndexType::redecl = "Index type `1` already declared."

DeclareIndexType::conflict = "Declaration of `1` as index type conflicts with
previous declaration."

HEPTensorSymbolQ::usage = "HEPTensorSymbolQ[t] returns True if t has been
declared as a tensor symbol with DeclareHEPTensor."

DeclareHEPTensor::usage = "DeclareHEPTensor[t, {i1, i2, ...}, f] declares a
tensor t with index list {i1, i2, ...}. The i1, i2 etc. must be
previously-defined index types and t must be a symbol or pattern. The argument f
is optional.  If t is a pattern and a matching expression is encountered in a
call to HEPTensorSignature, f is applied to the matching expression. The result
is discarded and the signature {i1, i2, ...} is returned. You may use f to
detect invalid arguments to tensor-valued functions and generate suitable
warnings.  You can also declare several tensors with the same signature with
DeclareHEPTensor[{t1, t2, ...}, {i1, i2, ...}, f]. Each of the t1, t2, ... may
be a symbol or a pattern."

DeclareHEPTensor::invidx = "Invalid index type in declaration of tensor `1`."

DeclareHEPTensor::redecl = "HEPTensor `1` already declared."

DeclareHEPTensor::conflict = "Declaration of tensor `1` with indices `2`
conflicts with previous declaration."

DeclareRealHEPTensor::usage = "DeclareRealHEPTensor[t, {i1, i2, ...}, f] is
equivalent to (DeclareHEPTensor[t, {i1, i2, ...}, f]; DeclareReal[t]). t may
be a symbol, pattern or a list of symbols and patterns. The argument f may be
ommitted."

IdentityHEPTensor::usage = "IdentityHEPTensor[t] represents the identity tensor
for index type t."

IdentityHEPTensorQ::usage = "IdentityHEPTensorQ[t] returns True if t was
declared as an identity HEPTensor (see DeclareIdentityHEPTensor)."

DeclareIdentityHEPTensor::usage = "DeclareIdentityHEPTensor[id, t] declares id
to be the identity matrix in the space of indices of type t. id is declared to
be real and have tensor signature {t, t}. TensorExpand will automatically remove
appearences of id by replacing contracted indices."

DeclareIdentityHEPTensor::invtpe = "Invalid index type `1`."

LinearQ::usage = "LinearQ[f] returns True if the function f was declared
as linear in all of its arguments (see DeclareLinear)."

DeclareLinear::usage = "DeclareLinear[f] declares the function f to be linear
in all of its arguments."

RealHolomorphicQ::usage = "RealHolomorphicQ[f] returns True if the function f
was declared real-holomorphic in all its arguments (see
DeclareRealHolomorphic)."

DeclareRealHolomorphic::usage = "DeclareRealHolomorphic[f] declares that the
function f is real-holomorphic in all its arguments. This means that
CC[f[z1, z2, ...]] is automatically replaced by f[CC[z1], CC[z2], ...]."

DeclareReal::usage = "DeclareReal[x] declares the expression x to be real."

RealQ::usage = "RealQ[x] returns True if the expression x was declared as
real (see DeclareReal)."

Scope::usage = "Scope[expr] evaluates the expression expr and makes
all declarations in expr local."

ScopeSymbol::usage = "ScopeSymbol[s] adds symbol s to the list of symbols
protected by Scope. At the end of a Scope the DownValues of s will be restored
to what they were at the beginning of Scope."

Attributes[Scope] = {HoldAll}

HEPTensorExpressionQ::usage = "HEPTensorExpressionQ[expr] returns True if expr is
a tensor-valued expression, i.e. if expr has suppressed tensor indices."

HEPTensorSignature::usage = "HEPTensorSignature[expr] returns the signature of
the tensor expression expr, i.e. the list of index types with which expr must be
evaluated."

HEPTensorSignature::conflict = "Incompatible tensors in tensor expression `1`."

HEPTensorSignature::invidx = "Invalid index list in tensor `1`."

Indices::usage = "Indices[expr] returns information about the free
and contracted indices of expression expr. It returns an expression of the
form {{if1->tf1, if2 -> tf2, ...}, {ic1 -> tc1, ic2 -> tc2, ...}}, where if1,
if2 etc are the free indices, ic1, ic2 etc the contracted indices and
tf1, tf2, ..., tc1, tc2, ... the corresponding index types."

Indices::mult = "Index repeated more than twice in product `1`."

Indices::sum = "Free indices do not match in sum `1`."

Indices::type = "Inconsistent type for index `1`."

RenameIndices::usage = "RenameIndices[expr] replaces all contracted indices
in expr with unique symbols."

HEPReplaceAll::usage = "HEPReplaceAll[expr, rules] applies rules to expr and
renames contracted indices in the replacements where necessary."

ExplicitIndices::usage = "ExplicitIndices[expr] makes all suppressed indices in
expression expr explicit by generating dummy indices. ExplicitIndices[expr, patt]
ignores sub-expressions which are free of pattern patt."

ExplicitIndices::suppidx = "ExplicitIndices only works on expressions with no
suppressed indices."

UserExplicitIndices::usage = "UserExplicitIndices is a function which can be
defined by the user. It is called by ExplicitIndices to handle index-free
notations introduced by the user. It is always called as
UserExplicitIndices[expr, patt]. See ExplicitIndices for an explanation of the
arguments."

HEPExpand::usage = "HEPExpand[expr, patt] performs a linear expansion of
an expression expr containing tensors, leaving subexpressions free of pattern
patt untouched. If the second argument is omitted, all subexpressions containing
tensors are expanded. Expressions without tensors are never expanded."

HEPDenominatorExpand::usage = "HEPDenominatorExpand[expr, patt] applies
HEPExpand to all denominators in expr, leaving subexpressions free of pattern
patt untouched. If the second argument is omitted, all subexpressions containing
tensors are expanded. Expressions without tensors are never expanded."

PullScalars::usage = "PullScalars[expr] pulls scalar factors out of the heads
of indexed tensor expressions, e.g. transforms (a t)[i] to a t[i], where a
is a scalar, t a tensor and i an index."

JoinIndexLists::usage = "JoinIndexLists[expr] joins the index lists of tensors
appearing in expr, i.e. replaces t[i1, i2,...][j1, j2,...] by
t[t1, i2, ..., j1, j2,...]. It also removes emtpy index lists like t[].
JoinIndexLists[expr, patt] only joins the index lists of tensors matching
pattern patt."

HEPContract::usage = "HEPContract[expr] removes appearences of identity tensors
in expr (as declared with DeclareIdentityHEPTensor or DeclareIndexType) by 
replacing indices appropriately."

HEPCollect::usage = "HEPCollect[expr, patt] factors out all expressions with
subexpressions matching pattern patt in expression expr. See also
HEPCollectListed."

Options[HEPCollect] = {CombineTensorStructures -> True}

HEPCollectListed::usage = "HEPCollectListed[expr, patt] works like
HEPCollect[expr, patt] but returns the result in the form {{t1, t2, ...},
 {c1, c2, ...}}, where the ti are the terms in expr matching patt and the ci are
the corresponding coefficients. HEPCollect[expr, patt] is equivalent to
Dot @@ HEPCollectListed[expr, patt]."

Options[HEPCollectListed] = {CombineTensorStructures -> True}

CombineTensorStructures::usage = "CombineTensorStructures is an option for
HEPCollect and HEPCollectListed. If set to True terms with the same tensor
structure but different names for the contracted indices are combined by
re-naming the contracted indices."

HEPCoefficient::usage = "HEPCoefficient[l, patt] returns the coefficient
of the term matching pattern patt in l, where l is a list of the form returned
by HEPCollectListed, i.e. of the form {{term1, term2, ...},
 {coeff1, coeff2, ...}}."

HEPCoefficient::mult = "Multiple matches found for pattern `1`"

HEPCoefficient::invarg = "Invalid argument list `1`"

HEPSelectTerms::usage = "HEPSelectTerms[clist, f] selects those terms from
clist for which f returns True. clist must be a list of the form
returned by HEPCollectListed, i.e. {{term1, term2, ...}, {coeff1, coeff2, ...}}.
The result is of the same form, but includes only those terms for which f
returns True."

HEPSelectTerms::invarg = "Invalid argument list `1`"

HEPGroupTerms::usage = "HEPGroupTerms[clist, f] groups the terms in clist
according to whether f applied to the term returns True or not. clist must be a
list of the form returned by HEPCollectListed. The result is a pair of lists
of the same form and is equivalent to {HEPSelectTerms[clist, f],
HEPSelectTerms[clist, f[#]=!=True &]}."

HEPGroupTerms::invarg = "Invalid argument list `1`"

CC::usage = "CC[expr] denotes the complex conjugate of expr. It automatically
distributes over sums, products, dot products and tensor indices."

SDot::usage = "SDot[t1, t2] denotes the full symmetric contraction of the
tensors t1 and t2. The two tensors must have the same tensor signature."

Sqr::usage = "Sqr[t] denotes the full contraction of the tensors t with
itself. It is equivalent to SDot[t,t]."

Dag::usage = "Dag[t] denotes the hermitian conjugate of the tensor t. It
is equivalent to the complex conjugate of t with a reversed index list."

Trans::usage = "Trans[t] denotes the transpose of the tensor t. It reverses
the index list of t."

HEPMultiply::usage = "HEPMultiply[ex1, ex2, ...] replaces dummy indices in ex1,
ex2 etc. by unique symbols and returns the product of the expressions."

HEPD::usage = "HEPD[expr, t[i, j, ...]] differentiates expr with
respect to the tensor t with indices i, j, etc. HEPD[expr, t1[i1, j1, ...],
t2[i2, j2, ...], ...] successively differentiates expr with respect to
t1, t2 etc. If HEPD cannot differentiate a subexpression it returns a
HEPDerivative expression. You can define rules for HEPDerivative to implement
derivatives of user-defined tensors or contractions."

HEPDerivative::usage = "HEPDerivative[expr, t[i, j, ...]] denotes the
derivative of expr with respect to tensor t with indices i, j, etc.
HEPDerivative expressions are returned by HEPD for subexpressions it cannot
differentiate. You can define rules for HEPDerivative to implement
derivatives of user-defined tensors or contractions." 

HEPIntegral::usage = 
"HEPIntegral[expr, {x1, x1min, x1max}, {x2, x2min, x2max}, ...] represents the
integral of expr over the variables x1, x2, ... in the limits (x1min, x1max),
 (x2min, x2max), etc. The Integral over x1 is the outermost integral and should
be done last. No automatic simplifications are performed on the
integral. HEPIntegral expressions are returned by the EliminateFeynmanParameter
fuction."

HEPSetZero::usage = "HEPSetZero[expr, patt] sets all tensors in expr matching
pattern patt to zero. Use this function in order to avoid generating expressions
like 0[i1, i2, ...]."


Begin["`Private`"]

(* declarations *)

DeclareIndexType[tt_List] := DeclareIndexType /@ tt

DeclareIndexType[t_, dim_, id_] := Which[
  (t === Plus || t === Times || t === NonCommutativeMultiply || t === Dot ||
   HEPTensorSymbolQ[t]) === True,
    Message[DeclareIndexType::conflict, t]; Abort[],
  IndexTypeQ[t] === True,
    Message[DeclareIndexType::redecl, t],
  True,
    IndexTypeQ[t] = True; IndexDimension[t] = dim;
    If[id=!=Null, DeclareIdentityHEPTensor[id, t]];]

DeclareIndexType[t_, dim_] := DeclareIndexType[t, dim, Null]

DeclareHEPTensor[tt_, ind_] := DeclareHEPTensor[tt, ind, Null]

DeclareHEPTensor[tt_List, ind_List, func_] := 
  Scan[DeclareHEPTensor[#, ind]&, tt]

DeclareHEPTensor[t_, ind_List, Null] := (HEPClearCache[]; Which[
  IndexTypeQ[t] === True,
    Message[DeclareHEPTensor::conflict, t, ind]; Abort[],
  And @@ (IndexTypeQ /@ ind) =!= True,
    Message[DeclareHEPTensor::invidx, t],
  (HEPTensorSymbolQ[t] && (HEPTensorSignature[t] =!= ind)) === True,
    Message[DeclareHEPTensor::conflict, t, ind]; Abort[],
  (HEPTensorSymbolQ[t] && (HEPTensorSignature[t] === ind)) === True,
    Message[DeclareHEPTensor::redecl, t],
  True, HEPTensorSignature[t] := ind;
        HEPTensorSymbolQ[t] := True
];)

DeclareHEPTensor[t_, ind_List, func_] := (HEPClearCache[]; Which[
  IndexTypeQ[t] === True,
    Message[DeclareHEPTensor::conflict, t, ind]; Abort[],
  And @@ (IndexTypeQ /@ ind) =!= True,
    Message[DeclareHEPTensor::invidx, t],
  (HEPTensorSymbolQ[t] && (HEPTensorSignature[t] =!= ind)) === True,
    Message[DeclareHEPTensor::conflict, t, ind]; Abort[],
  (HEPTensorSymbolQ[t] && (HEPTensorSignature[t] === ind)) === True,
    Message[DeclareHEPTensor::redecl, t],
  True, HEPTensorSignature[q:t] := (func[q];ind);
        HEPTensorSymbolQ[q:t] := (func[q];True)
];)

DeclareHEPTensor[t_, ind_, func_] := DeclareHEPTensor[t, {ind}, func]

DeclareRealHEPTensor[t_, ind_] :=
  (DeclareHEPTensor[t, ind]; DeclareReal[t])

DeclareRealHEPTensor[t_, ind_, func_] :=
  (DeclareHEPTensor[t, ind, func]; DeclareReal[t])

DeclareIdentityHEPTensor[id_, t_] := If[IndexTypeQ[t] =!= True,
  Message[DeclareIdentityHEPTensor::invtpe, t]; Abort[],
  HEPClearCache[];
  SetAttributes[id, Orderless]; 
  IdentityHEPTensorQ[id] = True;
  IdentityHEPTensor[t] = id;
  DeclareRealHEPTensor[id, {t,t}]]

DeclareLinear[f_List] := Scan[DeclareLinear, f]

DeclareLinear[f_] := (LinearQ[p:f] = True;)

DeclareLinear[f1_, ff__] := DeclareLinear[{f1, ff}]

DeclareRealHolomorphic[f_List] := Scan[DeclareRealHolomorphic, f]

DeclareRealHolomorphic[f_] := (RealHolomorphicQ[p:f] = True;)

DeclareRealHolomorphic[f1_, ff__] := DeclareRealHolomorphic[{f1, ff}]

DeclareReal[f_List] := Scan[DeclareReal, f]

DeclareReal[f_] := (RealQ[p:f] = True;)

DeclareReal[f1_, ff__] := DeclareReal[{f1, ff}]


DeclareReal[Pi]

scopeSymbols = {}

ScopeSymbols[s__] := (scopeSymbols = Union[scopeSymbols, {s}])

Scope[ex_] := Block[
  {dv = DownValues /@ scopeSymbols,
   oldScopeSymbols=scopeSymbols,
   result},

  result = ex;
  scopeSymbols = oldScopeSymbols;
  MapThread[(DownValues[#1] = #2)&, {scopeSymbols, dv}];
  result
]

ScopeSymbols[IndexTypeQ, HEPTensorSymbolQ, HEPTensorSignature, LinearQ,
 RealHolomorphicQ, IdentityHEPTensorQ, IndexDimension, IdentityHEPTensor]


(* Identity tensors *)

HEPTensorSignature[IdentityHEPTensor[t_]] := {t, t}

HEPTensorSymbolQ[_IdentityHEPTensor] = True

RealQ[_IdentityHEPTensor] = True

IdentityHEPTensorQ[_IdentityHEPTensor] = True


(* HEPTensorSignature *)

tensorQ[t_] := (HEPTensorSignature[t] =!= {})

HEPTensorSignature[(t_?tensorQ)[ii___]] :=
  Block[{sig=HEPTensorSignature[t]},
    If[Length[sig]<Length[{ii}],
       Message[HEPTensorSignature::invidx, t[ii]]; Abort[],
       Take[sig,Length[{ii}]-Length[sig]]
    ]
  ]

HEPTensorSignature[HoldPattern[Plus[tt__]]] :=
Block[
  {sig = HEPTensorSignature /@ {tt}},

  If[Equal @@ sig =!= True,
     Message[HEPTensorSignature::conflict, Plus[tt]]; Abort[],
     First[sig]]
]

HEPTensorSignature[t:((Times|NonCommutativeMultiply)
  [___,_?tensorQ,___,_?tensorQ,___])] := 
  (Message[HEPTensorSignature::conflict, t]; Abort[])

HEPTensorSignature[(h:(Times|NonCommutativeMultiply))[a___,t_?tensorQ,b___]] :=
  HEPTensorSignature[t]

flagIndexTypes[l_] := Block[{res=l},
  res = FixedPoint[Replace[#, {a___, t_, b___, t_, c___} ->
    {a, left[t], b, right[t], c}]&, res];
  Replace[res, t:Except[_left|_right] -> single[t], {1}]
]

HEPTensorSignature[HoldPattern[Dot[tt__]]] := Block[
  {types, sig, left, right, single},

  sig = HEPTensorSignature /@ {tt};
  types = Union @@ sig;
  Outer[
    If[Count[#1, #2] > 2,
      Message[HEPTensorSignature::conflict, Dot[tt]]; Abort[]]&,
    sig, types];
  sig = Join @@ (flagIndexTypes /@ sig);
  Scan[
    If[Count[sig, single[#]] > 2, 
      Message[HEPTensorSignature::conflict, Dot[tt]]; Abort[]]&,
    types];
  sig = sig //. {x___, right[t_], y___, left[t_], z___} :>
    {x, y, z} /; FreeQ[{y}, t];
  sig = sig //. {x___, single[t_], y___, left[t_], z___} :>
    {x, y, z} /; FreeQ[{x}, t];
  sig = sig //. {x___, right[t_], y___, single[t_], z___} :>
    {x, y, z} /; FreeQ[{z}, t];
  sig = sig //. {x___, single[t_], y___, single[t_], z___} :> {x, y, z};
  If[MatchQ[sig, {___, single[t_], ___, (single|left|right)[t_], ___}],
    Message[HEPTensorSignature::conflict, Dot[tt]]; Abort[]];
  If[MatchQ[sig, {___, (single|left|right)[t_], ___, single[t_], ___}],
    Message[HEPTensorSignature::conflict, Dot[tt]]; Abort[]];
  sig /. single|left|right -> Identity
]

HEPTensorSignature[(t_?tensorQ)^k_] :=
  (Message[HEPTensorSignature::conflict, t^k]; Abort[])

HEPTensorSignature[SDot[t1_, t2_]] := Block[{sig1},
  sig1 = HEPTensorSignature[t1];
  If[HEPTensorSignature[t2] === sig1, {},
    Message[HEPTensorSignature::conflict, SDot[t1, t2]]; Abort[]]
]

HEPTensorSignature[Sqr[t_]] := (HEPTensorSignature[t]; {})

HEPTensorSignature[CC[t_]] := HEPTensorSignature[t]

HEPTensorSignature[Dag[t_]] := Reverse[HEPTensorSignature[t]]

HEPTensorSignature[Trans[t_]] := Reverse[HEPTensorSignature[t]]

HEPTensorSignature[HEPDerivative[ex_, t_]] := HEPTensorSignature[ex]

HEPTensorSignature[x_] = {}


(* Indices *)

mult::usage = "mult is a private symbol representing a multilinear function."

contractIndices[if_List, ic_List] := Block[{jf, jc},
  {jf, jc} =
    FixedPoint[Replace[#,{{a___,i_,b___,i_,c___},{d___}} :> 
                         {{a,b,c},{d,i}}, {0}]&, {if,ic}];
  Replace[jf, {___, i_ -> _, ___, i_ -> _, ___} :>
    (Message[Indices::type, i]; Abort[])];
  Replace[jc, {___, i_ -> _, ___, i_ -> _, ___} :>
    (Message[Indices::type, i]; Abort[])];
  {Union[jf], Union[jc]}
]

getIndices[mult[t_]] := getIndices[t]

getIndices[mult[]] = {{},{}}

getIndices[(m:(mult|Times|(_?LinearQ)))[t1_,tt___]] := Block[
  {if1, if, ic1, ic, result},

  {if1, ic1} = getIndices[t1];
  {if, ic} = getIndices[mult[tt]];
  If[MatchQ[{ic1, ic}, {{___,i_,___},{___,i_,___}}],
     Message[Indices::mult, m[t1,tt]]; Abort[]];
  contractIndices[Join[if1,if], Join[ic1,ic]]
]

getIndices[HoldPattern[Plus[tt__]]] := Block[
  {inds, if, ic},

  inds = getIndices /@ {tt};
  if = First /@ inds;
  ic = Last /@ inds;
  If[SameQ @@ if, {First[if], Union@@ic},
     Message[Indices::sum, Plus[tt]]; Abort[]]
]

getIndices[CC[x_]] := getIndices[x]

getIndices[Sqr[ex_]] := Block[{inds = getIndices[ex]},
  If[inds[[2]] === {}, {{},inds[[1]]}, Message[Indices::mult, Sqr[ex]]; Abort[]]
]

getIndices[ex_^2] := Block[{inds = getIndices[ex]},
  If[inds[[2]] === {}, {{},inds[[1]]}, Message[Indices::mult, Sqr[ex]]; Abort[]]
]

getIndices[ex_^k_] := If[getIndices[ex] === {{},{}}, {{},{}},
  Message[Indices::mult, ex^k]; Abort[]]

getIndices[(t_?tensorQ)[ii__]] := Block[
  {sig=HEPTensorSignature[t], n=Length[{ii}]},

  contractIndices[MapThread[Rule, {{ii}, Take[sig, n]}],{}]
]

getIndices[HEPDerivative[ex_, t_]] := Block[{f1, c1, f2, c2},
  {f1, c1} = getIndices[ex];
  {f2, c2} = getIndices[t];
  {Join[f1, f2], Join[c1, c2]}
]

getIndices[_] = {{},{}}

Indices[ex_] := Block[{$RecursionLimit=Infinity},
  HEPTensorSignature[ex]; Sort /@ getIndices[ex]]


(* RenameIndices *)

RenameIndices[ex_] := Block[{dummies},
  dummies = Indices[ex][[2]] /. (i_ -> _) :> (i -> Unique[]);
  ex /. dummies
]


(* HEPReplaceAll *)

pow::usage = "pow is an internal symbol equivalent to Power." 

HEPReplaceAll[ex_, rules_] := Block[{newrules},
  newrules = rules /. (Rule|RuleDelayed)[lhs_, rhs_] ->
    (lhs :> RenameIndices[rhs]);
  ex /. newrules
]

HEPReplaceAll[ex_, rules_, levelspec_] := Block[{newrules},
  newrules = rules /. (Rule|RuleDelayed)[lhs_, rhs_] ->
    (lhs :> RenameIndices[rhs]);
  ReplaceAll[ex, newrules, levelspec]
]


(* ExplicitIndices *)

putIndices[ex_, patt_] := ex /; FreeQ[ex, patt]

putIndices[(h:(Plus|Times|NonCommutativeMultiply))[aa__], patt_] :=
  h @@ (putIndices[#, patt]& /@ {aa})

putIndices[d:(_Dot)[ii___], patt_] := Block[
  {aa, inds, ind, sigext, ifirst, isecond, index, indexl, indexr},

  aa = List @@ Head[d];
  sigext = HEPTensorSignature[Dot @@ aa];
  extind = Transpose[{sigext,{ii}}];
  inds = Map[index, HEPTensorSignature /@ aa, {2}] //.
    {i1___,index[t_],i2___,index[t_],i3___} -> {i1,indexl[t],i2,indexr[t],i3};

  (* generate indices *)
  inds = inds //. {
    {a___,{i1___,indexr[t_],i2___},Shortest[b___],
      {j1___,indexl[t_],j2___},c___} :> 
      (ind = Unique[]; {a,{i1,index[t,ind],i2},b,{j1,index[t,ind],j2},c}),
    {a___,{i1___,index[t_],i2___},Shortest[b___],
      {j1___,indexl[t_],j2___},c___} :> 
      (ind = Unique[]; {a,{i1,index[t,ind],i2},b,{j1,index[t,ind],j2},c}),
    {a___,{i1___,indexr[t_],i2___},Shortest[b___],
      {j1___,index[t_],j2___},c___} :> 
      (ind = Unique[]; {a,{i1,index[t,ind],i2},b,{j1,index[t,ind],j2},c}),
    {a___,{i1___,index[t_],i2___},Shortest[b___],
      {j1___,index[t_],j2___},c___} :> 
      (ind = Unique[]; {a,{i1,index[t,ind],i2},b,{j1,index[t,ind],j2},c})
  };

  (* assign external matrix indices *)
  {inds, extind} = {inds, extind} //. {
    {{a___,{i1___,indexl[t_],i2___},b___,{j1___,indexr[t_],j2___},c___},
      {t1___,{t_,kl_},t2___,{t_,kr_},t3___}} ->
      {{a,{i1,index[t,kl],i2},b,{j1,index[t,kr],j2},c}, {t1,t2,t3}},
    {{a___,{i1___,indexl[t_],i2___,indexr[t_],i3___},b___},
      {t1___,{t_,kl_},t2___,{t_,kr_},t3___}} ->
      {{a,{i1,index[t,kl],i2,index[t,kr],i3},b}, {t1,t2,t3}}
  };

  (* assign external single indices *)
  {inds, extind} = {inds, extind} //. 
    {{a___,{i1___,(index|indexr|indexl)[t_],i2___},b___},
      {t1___,{t_,k_},t2___}} ->
      {{a,{i1,index[t,k],i2},b}, {t1,t2}};

  Times @@
    Replace[MapThread[(#1 @@ #2)&, {aa, inds /. index[t_,i_]->i}],
      (h_?tensorQ)[jj__][kk__] -> h[jj,kk], {1}]
]

putIndices[d_Dot, patt_] := putIndices[d[], patt]

putIndices[SDot[a_, b_], patt_] := Block[{n=Length[HEPTensorSignature[a]]},
  inds = Table[Unique[], {n}];
  If[n === 0, a b, (a @@ inds) (b @@ inds)]
]

putIndices[Sqr[a_], patt_] := Block[{n=Length[HEPTensorSignature[a]]},
  inds = Table[Unique[], {n}];
  If[n === 0, a^2, (a @@ inds)^2]
]

putIndices[x_ , patt_] := UserExplicitIndices[x, patt]

ExplicitIndices[ex_, patt_] := If[tensorQ[ex],
  Message[ExplicitIndices::suppidx]; Abort[], putIndices[ex, patt]]

ExplicitIndices[ex_] := ExplicitIndices[ex, _]


(* UserExplicitIndices *)

UserExplicitIndices[ex_, patt_] = ex


(* HEPExpand *)

DeclareLinear[sdot]

subscriptExpand[t_[ii__], patt_] := t[ii] /; FreeQ[t, patt]

subscriptExpand[(a_ (t_?tensorQ))[ii___], patt_] :=
  a subscriptExpand[t[ii], patt]

subscriptExpand[(t_Plus)[ii___], patt_] := subscriptExpand[#[ii], patt]& /@ t

subscriptExpand[t_[ii___], patt_] := t[ii]

HEPExpand[ex_] := HEPExpand[ex, _?tensorQ]

HEPExpand[ex_, patt_] := ex /; FreeQ[ex, patt]

HEPExpand[l_List, patt_] := HEPExpand[#, patt]& /@ l

HEPExpand[(t_?tensorQ)[ii___], patt_] := 
  subscriptExpand[HEPExpand[t, patt][ii], patt]

HEPExpand[s_Plus, patt_] := HEPExpand[#, patt]& /@ s

HEPExpand[p_Times, patt_] := Expand[HEPExpand[#, patt]& /@ p, patt]

HEPExpand[ex:(_?LinearQ)[__], patt_] :=
  tensorExpand[HEPExpand[#,patt]& /@ ex, patt]

HEPExpand[Sqr[ex_], patt_] :=
  (HEPExpand[sdot[ex, ex], patt] /. sdot -> SDot) /; tensorQ[ex]

HEPExpand[Sqr[ex_], patt_] := ex^2 /; !tensorQ[ex]

HEPExpand[ex_^k_Integer, patt_] := Expand[HEPExpand[ex, patt]^k, patt] /; k > 0

HEPExpand[HEPIntegral[i_, x__], patt_] := Block[{vp},
  vp = Alternatives@@(First /@ {x});
  diffIntExpand[HEPIntegral[HEPExpand[i, patt], x], patt, vp]
]

HEPExpand[HEPDerivative[i_, (t_?tensorQ)[ii__]], patt_] := 
  diffIntExpand[HEPDerivative[HEPExpand[i, patt], t[ii]], patt, t]

HEPExpand[x_, patt_] = x

diffIntExpand[i_, patt_, vp_] := i /; FreeQ[i, patt]

diffIntExpand[(h:(HEPIntegral|HEPDerivative))[s_Plus, x__], patt_, vp_] :=
  diffIntExpand[h[#, x], patt, vp]& /@ s

diffIntExpand[(h:(HEPIntegral|HEPDerivative))[a_, x__], _, vp_] :=
  a h[1, x] /; FreeQ[a, vp]

diffIntExpand[(h:(HEPIntegral|HEPDerivative))[a_ b_, x__], patt_, vp_] :=
  a diffIntExpand[h[b, x], patt, vp] /; FreeQ[a, vp]

diffIntExpand[i_, _, _] := i

tensorExpand[t_, patt_] := t /; FreeQ[t, patt]

tensorExpand[(h_?LinearQ)[a___,s_Plus,b___], patt_] :=
  tensorExpand[h[a,#,b], patt]& /@ s /; !FreeQ[s, patt]

tensorExpand[(h_?LinearQ)[x___, p_Times, y___], patt_] := Block[{fac,tens},
  {tens,fac} = unZip[List@@p, tensorQ];
  If[Length[tens] > 1, Message[HEPTensorSignature::conflict, tens]; Abort[]];
  Expand[(Times@@fac) tensorExpand[h[x,Sequence@@tens,y], patt], patt]
] /; !FreeQ[p, patt]

tensorExpand[x_, patt_] = x


(* HEPDenominatorExpand *)

HEPDenominatorExpand[ex_] := HEPDenominatorExpand[ex, _?tensorQ]

HEPDenominatorExpand[s_Plus, patt_] := HEPDenominatorExpand[#, patt]& /@ s

HEPDenominatorExpand[p_Times, patt_] := HEPDenominatorExpand[#, patt]& /@ p

HEPDenominatorExpand[d_^k_Integer, patt_] :=
  HEPExpand[d^(-k), patt]^(-1) /; k < 0

HEPDenominatorExpand[x_, patt_] := x


(* PullScalars *)

subscriptPull[(a_ (t_?tensorQ))[ii___]] := a subscriptPull[t[ii]]

subscriptPull[t_[ii___]] := t[ii]

PullScalars[ex_, patt_] := ex /; FreeQ[ex, patt]

PullScalars[(t_?tensorQ)[ii___], patt_] :=
  subscriptPull[PullScalars[t, patt][ii]]

PullScalars[s_Plus, patt_] := PullScalars[#, patt]& /@ s

PullScalars[p_Times, patt_] := PullScalars[#, patt]& /@ p

PullScalars[(h_?LinearQ)[x__], patt_] :=
  pullScalars[h @@ (PullScalars[#, patt]& /@ {x})]

PullScalars[x_, patt_] = x

PullScalars[x_] := PullScalars[x, _]

pullScalars[(h_?LinearQ)[x___, a_ (t_?tensorQ), y___]] :=
  a PullScalars[h[x,t,y]]

pullScalars[x_] = x


(* JoinIndexLists *)

JoinIndexLists[ex_, patt_] := ex //. {
  (t:patt)[] :> t /; tensorQ[t],
  (t:patt)[ii___][jj___] :> t[ii,jj] /; tensorQ[t]
}

JoinIndexLists[ex_] := JoinIndexLists[ex, _]


(* HEPContract *)

HEPContract[ex_, patt_] := Block[{res=ex, times, dot},
  res = JoinIndexLists[res, patt];
  res = res /. 
    (id:patt)[i_Integer,j_Integer] :> If[i===j, 1, 0] /; IdentityHEPTensorQ[id];
  res = res /. Times -> times;
  res = res //. {
    times[a___,(id:patt)[i_,j_], b___] :> (times[a,b] /. j -> i) /;
      IdentityHEPTensorQ[id] && !FreeQ[times[a,b], j],
    times[a___,(id:patt)[i_,j_], b___] :> (times[a,b] /. i -> j) /;
      IdentityHEPTensorQ[id] && !FreeQ[times[a,b], i]
  };
  res = res /. times -> Times;
  res = res /. 
    (id:patt)[i_,i_] :> IndexDimension[First[HEPTensorSignature[id]]] /;
      IdentityHEPTensorQ[id];
  res = res /. 
    (id:patt)[i_,j_]^2 :> IndexDimension[First[HEPTensorSignature[id]]] /;
      IdentityHEPTensorQ[id];
  res = res /. Dot -> dot;
  res = res //. 
    dot[a___,id:patt,b___] :> dot[a,b] /; IdentityHEPTensorQ[id];
  res = res /. dot[a_, b_] :> Block[{sig=HEPTensorSignature[a]},
    If[HEPTensorSignature[b] === sig && Union[sig] === Sort[sig],
      SDot[a, b],
      dot[a, b]
    ]
  ];
  res = res /. dot -> Dot;
  res
]

HEPContract[ex_] := HEPContract[ex, _]


(* HEPCollect *)

sumToList[s_Plus] := List@@s

sumToList[s_] := {s}

prodToList[p_Times] := List@@p

prodToList[p_] := {p}

unZip[l_List, f_] := Block[{true={},false={}},
  Scan[If[f[#]===True, AppendTo[true, #], AppendTo[false, #]]&, l];
  {true,false}
]

HEPCollectListed[ex_, patt_, OptionsPattern[]] := Block[
  {lst = prodToList /@ sumToList[HEPExpand[ex, patt]],
   sel = !FreeQ[#, patt]&, terms, coeffs},

  lst = unZip[#, sel]& /@ lst;
  lst = Apply[Times, lst, {2}];
  terms = Union[First /@ lst];
  coeffs = (Plus@@Cases[lst, {#, c_} -> c])& /@ terms;
  If[OptionValue[CombineTensorStructures],
    Transpose[combineTensorStructures[{}, Transpose[{terms, coeffs}]]],
    {terms, coeffs}]
]

HEPCollect[ex_, patt_, opt:OptionsPattern[]] :=
  Dot @@ HEPCollectListed[ex, patt, opt]


(* combineHEPTensors *)

Quiet[
  indexToPattern[ex_, i_Symbol] := {ex /. i -> i_, {}},
  RuleDelayed::rhs]

Quiet[
  indexToPattern[ex_, j_] := Module[{i}, {ex /. j -> i_, {j -> i}}],
  RuleDelayed::rhs]

indicesToPattern[ex_, ii_List] := Block[{res=ex, rul={}},
  Do[
    res = indexToPattern[res, i];
    rul = Join[rul, Last[res]];
    res = First[res],
    {i, ii}
  ];

  {res, rul}
]

combineTensorStructures[res_, {}] := res

combineTensorStructures[res_, {tc_}] := Append[res, tc]

combineTensorStructures[res_, {{t1_, c1_}, tc__}] := Block[
  {tt, cc, indices, repl, patt, pattind, match, nomatch, ind1},

  tt = First /@ {tc};
  cc = Last /@ {tc};
  indices = First /@ Intersection[First[Indices[t1]], First[Indices[c1]]];
  {patt, repl} = indicesToPattern[t1, indices];
  pattind = indices /. repl;
  {match, nomatch} = unZip[{tc}, MatchQ[#, {patt, _}]&];

  If[match =!= {},
    match = Append[match, {t1, c1}];
    newind = Table[Unique[], {Length[pattind]}];
    match =
      (Last[#] /. MapThread[Rule, {First[#] /. patt -> pattind, newind}])& /@
      match;
    newt1 = t1 /. MapThread[Rule, {indices, newind}];
    combineTensorStructures[Append[res, {newt1, Plus @@ match}], nomatch]
    , (* else *)
    combineTensorStructures[Append[res, {t1, c1}], nomatch]
  ]
]


(* HEPCoefficient *)

HEPCoefficient[{}, _] = 0

HEPCoefficient[{terms_List, coeff_List}, patt_] := Block[{pos, res},
  pos = Position[terms, patt];
  Which[
    pos === {},        0,
    Length[pos] === 1, Extract[coeff, First[pos]],
    True,              Message[HEPCoefficient::mult, patt]; Abort[]]
]

HEPCoefficient[aa___] := (Message[HEPCoefficient::invarg, {aa}]; Abort[])


(* HEPSelectTerms *)

HEPSelectTerms[{}, _] = {}

HEPSelectTerms[t:{_List, _List}, f_] := Block[{res},
  res = Select[Transpose[t], f[First[#]]&];
  If[res =!= {}, res = Transpose[res]];
  res
]

HEPSelectTerms[aa___] := (Message[HEPSelectTerms::invarg, {aa}]; Abort[])


(* HEPGroupTerms *)

HEPGroupTerms[{}, _] = {{},{}}

HEPGroupTerms[t:{terms_List, coeff_List}, f_] := Block[
  {tt=Transpose[t], yes={}, no={}},
  Scan[If[f[First[#]]===True, AppendTo[yes, #], AppendTo[no, #]]&, tt];
  yes = If[yes =!= {}, Transpose[yes], {{},{}}];
  no = If[no =!= {}, Transpose[no], {{},{}}];
  {yes, no}
]

HEPGroupTerms[aa___] := (Message[HEPGroupTerms::invarg, {aa}]; Abort[])



(* CC *)

DeclareReal[{_Re, _Im}]

CC[x_?NumberQ] := Conjugate[x]

CC[CC[x_]] := x

CC[x_^(n_)] := CC[x]^n

CC[s:((_?RealHolomorphicQ)[__])] := CC /@ s

CC[(t_?tensorQ)[ii___]] := CC[t][ii]

CC[Sqr[x_]] := Sqr[CC[x]]

CC[x_?RealQ] = x


(* Dag and Trans *)

Dag[Trans[t_]] := CC[t]

Trans[Dag[t_]] := CC[t]

CC[Dag[t_]] := Trans[t]

Dag[CC[t]] := Trans[t]

CC[Trans[t]] := Dag[t]

Trans[CC[t]] := Dag[t]

DeclareLinear[{Dag, Trans}]

Dag[t_?tensorQ][ii__] := CC[t]@@Reverse[{ii}] /;
  Length[{ii}] === Length[HEPTensorSignature[t]]

Trans[t_?tensorQ][ii__] := t@@Reverse[{ii}] /;
  Length[{ii}] === Length[HEPTensorSignature[t]]


(* HEPMultiply *)

HEPMultiply[a___] := Times @@ ( RenameIndices /@ {a})

DeclareLinear[{NonCommutativeMultiply,Dot}]

DeclareRealHolomorphic[{Plus, Times, NonCommutativeMultiply, Dot}]


(* SDot *)

Attributes[SDot] = {Orderless}

DeclareLinear[SDot]

DeclareRealHolomorphic[SDot]

SDot[0, _] = 0

SDot[t_, t_] := Sqr[t]


(* Sqr *)

Sqr[n_?NumericQ] := n^2


(* HEPD *)

HEPD[ex_, t1_, tt__] := HEPD[HEPD[ex, t1], tt]

HEPD[ex_, t_?tensorQ] := (Message[HEPD::isup, t]; Abort[])

HEPD[ex_, x:Except[(_?tensorQ)[__]]] := D[ex, x]

HEPD[ex_, t_[__]] := 0 /; FreeQ[ex, t]

HEPD[ex_Plus, t_[ii__]] := HEPD[#, t[ii]]& /@ ex

HEPD[HoldPattern[Times[aa__]], t_[ii__]] := Block[{fac={aa}},
  Sum[Times@@Join[Take[fac, k-1], {HEPD[fac[[k]], t[ii]]}, Drop[fac, k]],
    {k, Length[fac]}]
]

HEPD[b_^n_, t_[ii__]] := n b^(n-1) HEPD[b, t[ii]] /; FreeQ[n, t]

HEPD[CC[ex_], t_[ii__]] := CC[HEPD[ex, t[ii]]]

HEPD[t_[ii__], t_[jj__]] := Block[{inds=HEPTensorSignature[t]},
  Times@@MapThread[IdentityHEPTensor[#1][#2, #3]&, {inds, {ii}, {jj}}]
] /; Length[{ii}] === Length[{jj}]

HEPD[s_, t_[ii__]] := Block[{exp},
  exp=HEPExpand[s, t];
  If[exp === s,
    HEPDerivative[s, t[ii]],
    HEPD[exp, t[ii]]
  ]
]


HEPDerivative[SDot[s_, t_], t_[ii__]] := s[ii] /; FreeQ[s, t]

HEPDerivative[Sqr[t_], t_[ii__]] := 2 t[ii]


(* HEPIntegral *)

HEPIntegral[i_] := i


(* HEPSetZero *)

HEPSetZero[ex_, patt_] := (ex //. patt[__] -> 0) /. patt -> 0



End[]

EndPackage[]

