PrependTo[$Path, "../.."]

Needs["HEPMath`"]

Needs["HEPMath`Compile`"]

Needs["CCompilerDriver`"]

ltresults = Get["ltresults.mo"]

approxEqual[a_, b_] := Abs[a-b] < Max[10.0^(-14)*Max[Abs[a],Abs[b]], 10.0^(-14)]

DeclareLorentzVectors[p,q,k,l,e1,e2]


testHEPPrepare[] := Block[{tests, mu, m, expr, testExpr},
  mu = 172.0;

  expr = {
    PaVe[1,mu,0][0][Sqr[p]]*Den[p, m] +
      PaVe[2,mu,0][1][Sqr[p], SDot[p,q], SDot[p,q]],
    PaVe[2,mu,0][0][Sqr[p], 0, 0],
    PaVe[2,mu,0][0][Sqr[p], 10.0^(-14), 0],
    PaVe[2,mu,1][0][Sqr[p], 0, 0],
    PaVe[2,mu,2][0][Sqr[p], 0, 0]
  };

  testExpr[ex_, res_] := Block[{cex=ex},
    cex = HEPCompile[{{p, _Real, 1}, {q, _Real, 1}, {m, _Real}}, HEPPrepare[cex],
                     CompilationOptions->{"ExpressionOptimization"->True,
                                          "InlineCompiledFunctions"->True,
                                          "InlineExternalDefinitions"->True}];
    cex = cex[{400., 0., 0., 0.}, {25., 0., 0., 0.}, 15.0];
    approxEqual[cex, res]
  ];

  tests = MapThread[testExpr, {expr, Take[ltresults, Length[expr]]}];
  And @@ tests
]

testHEPCompile[] := Block[
  {tests={}, p={10., 0.3, 0.6, 1.0}, e1, e2, cex},
  cex = HEPCompile[{{sg1, _Integer}, {sg2, _Integer}},
    Block[{e1=Pol[p][sg1], e2=CC[Pol[p]][sg2]}, HEPPrepare[SDot[e1, e2]]],
    CompilationOptions->{"ExpressionOptimization"->True,
                         "InlineCompiledFunctions"->True,
                         "InlineExternalDefinitions"->True}];
  AppendTo[tests, approxEqual[cex[-1,-1], -1.0]];
  AppendTo[tests, approxEqual[cex[-1,+1],  0.0]];
  AppendTo[tests, approxEqual[cex[+1,-1],  0.0]];
  AppendTo[tests, approxEqual[cex[+1,+1], -1.0]];

  And @@ tests  
]


testHEPCodeGenerate[] := Block[
  {expr,x,y,b0,cf,p,q,k,l,cfuncs={},fnames={},pdfid,builddir,libname,modules},

  expr = x^2 + y^2;
  cf = HEPCompile[{{x, _Real}, {y, _Real}}, HEPEvaluate[expr] + 1.0];
  AppendTo[cfuncs, cf];
  AppendTo[fnames, "f1"];

  If[HEPMath`Config`$LoopToolsLib =!= "",
    b0 = PaVe[2, 1.0, 0][0][Sqr[p], m^2, m^2];
    cf = HEPCompile[{{p, _Real, 1}, {m, _Real}}, HEPPrepare[b0]];
    AppendTo[cfuncs, cf];
    AppendTo[fnames, "f2"]
  ];

  cf = HEPCompile[{{p, _Real, 1}, {q, _Real, 1}, {k, _Real, 1}, {l, _Real, 1}},
    HEPPrepare[EpsDot[p, q, k, l]]];
  AppendTo[cfuncs, cf];
  AppendTo[fnames, "f3"];

  If[HEPMath`Config`$LHAPDFLib =!= "",
    pdfid = LHAPDFOpen["cteq6l1", 0];
    LHAPDFClose[pdfid];
    cf = HEPCompile[{{x, _Real}},
      Block[{xf},
        LHAPDFOpenID[HEPEvaluate[pdfid]];
        xf = LHAPDF[HEPEvaluate[pdfid], 21, x, 100.0];
        LHAPDFClose[HEPEvaluate[pdfid]];
        xf
    ]];
    AppendTo[cfuncs, cf];
    AppendTo[fnames, "f4"]
  ];

  builddir = "build";
  libname = "myfuncs";
  If[DirectoryQ[builddir], DeleteDirectory[builddir, DeleteContents->True]];
  CreateDirectory[builddir];

  modules = HEPCodeGenerate[cfuncs, fnames, libname, TargetDirectory->builddir];
  CreateLibrary[modules, "_"<>libname,
    "TargetDirectory" -> builddir,
    "Debug"->True, "Libraries"->{"WolframRTL", "dl"},
    "ShellCommandFunction"->Print];
  DeleteDirectory[#, DeleteContents->True]& /@
    FileNames[FileNameJoin[{builddir, "Working*"}]];
]


fail = False;

WriteString[$Output, "testHEPPrepare: "<>If[testHEPPrepare[],"PASS\n", fail=True; "FAIL\n"]];
WriteString[$Output, "testHEPCompile: "<>If[testHEPCompile[],"PASS\n", fail=True; "FAIL\n"]];
testHEPCodeGenerate[];

If[fail, Exit[1], Exit[]];

